(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Reduction_Pair_Implementations
imports
  Matrix_Poly
  Argument_Filter
  Poly_Order
  Efficient_RPO
  SCNP_Impl
  List_Order_Implementations
  KBO_Impl
  "$AFP/Sqrt_Babylonian/Sqrt_Babylonian"
begin

datatype 'f redtriple_impl = 
    Int_carrier "('f, int) lpoly_interL" 
  | Int_nl_carrier "('f, int) poly_inter_list" 
  | Rat_carrier "('f, rat) lpoly_interL"
  | Rat_nl_carrier rat "('f, rat) poly_inter_list"
  | Mini_alg_carrier "('f, mini_alg_number) lpoly_interL"
  | Mini_alg_nl_carrier mini_alg_number "('f, mini_alg_number) poly_inter_list"
  | Arctic_carrier "('f, arctic) lpoly_interL" 
  | Arctic_rat_carrier "('f, rat arctic_delta) lpoly_interL" 
  | Int_mat_carrier nat nat "('f, int mat) lpoly_interL"
  | Rat_mat_carrier nat nat "('f, rat mat) lpoly_interL"
  | Mini_alg_mat_carrier nat nat "('f, mini_alg_number mat) lpoly_interL"
  | Arc_mat_carrier nat "('f, arctic mat) lpoly_interL"
  | Arc_rat_mat_carrier nat "('f, rat arctic_delta mat) lpoly_interL"
  | RPO "'f status_prec_repr" "'f afs_list"
  | KBO "'f prec_weight_repr" "'f afs_list"

definition faulty_redtriple :: "string \<Rightarrow> ('f :: show, 'v :: show)redtriple"
where "faulty_redtriple s \<equiv> \<lparr>redtriple.valid = error (shows_string s), s = \<lambda> _. succeed, ns = \<lambda> _. succeed, nst = \<lambda> _. succeed, af = full_af, mono = \<lambda> _. succeed, desc = id, cpx = no_complexity \<rparr>"

lemma faulty_redtriple: "generic_redtriple_impl faulty_redtriple"
  by (unfold_locales, unfold faulty_redtriple_def, auto)

definition faulty_non_inf_order :: "string \<Rightarrow> 'a \<Rightarrow> ('f :: show, 'v :: show)non_inf_order"
where "faulty_non_inf_order s F \<equiv> \<lparr>non_inf_order.valid = error (shows_string s), ns = \<lambda> _. succeed, cc = \<lambda> _. succeed, af = \<lambda> _ _. Wild, desc = id\<rparr>"

lemma faulty_non_inf_order: "generic_non_inf_order_impl faulty_non_inf_order"
  by (unfold_locales, unfold faulty_non_inf_order_def, auto)

fun filter_prec_weight_repr :: "'f afs \<Rightarrow> 'f prec_weight_repr \<Rightarrow> 'f filtered prec_weight_repr"
where "filter_prec_weight_repr \<pi> (prw,w0) = (
   let fprw = filter (\<lambda> (fn,e). case \<pi> fn of Collapse _ \<Rightarrow> False | AFList _ \<Rightarrow> True) prw;
       mprw = map (\<lambda> ((f,n),e). ((FPair f n, case \<pi> (f,n) of AFList l \<Rightarrow> length l | Collapse _ \<Rightarrow> 0),e)) fprw
    in (mprw,w0))"

fun prec_repr_to_pr :: "'f status_prec_repr \<Rightarrow> (('f :: key) filtered \<Rightarrow> nat \<Rightarrow> nat)" where
  "prec_repr_to_pr prs = (let m = ceta_map_of prs in (\<lambda> fp n. case fp of 
    FPair f a \<Rightarrow> case m (f, a) of None \<Rightarrow> 0 | Some x \<Rightarrow> fst x))" 

fun prec_repr_to_status :: "'f status_prec_repr \<Rightarrow> (('f :: key) filtered \<Rightarrow> nat \<Rightarrow> order_tag)" where
  "prec_repr_to_status prs = (let m = ceta_map_of prs in (\<lambda> fp n. case fp of
    FPair f a \<Rightarrow> case m (f, a) of None \<Rightarrow> Lex | Some x \<Rightarrow> snd x))" 

fun get_redtriple :: "('f :: {key,show}) redtriple_impl \<Rightarrow> ('f,'v :: show)redtriple"
where 
  "get_redtriple (Int_carrier I) = int_redtriple I"
| "get_redtriple (Int_nl_carrier I) = create_nlpoly_redtriple succeed (1 :: int) (op >) True True I"
| "get_redtriple (Rat_carrier I) = delta_redtriple 1 I"
| "get_redtriple (Rat_nl_carrier d I) = delta_nl_redtriple d I"
| "get_redtriple (Mini_alg_carrier I) = delta_redtriple 1 I"
| "get_redtriple (Mini_alg_nl_carrier d I) = delta_nl_redtriple d I"
| "get_redtriple (Arctic_carrier I) = arctic_redtriple I"
| "get_redtriple (Arctic_rat_carrier I) = arctic_delta_redtriple I"
| "get_redtriple (Int_mat_carrier n sd I) = int_mat_redtriple n sd I"
| "get_redtriple (Rat_mat_carrier n sd I) = delta_mat_redtriple n sd 1 I"
| "get_redtriple (Mini_alg_mat_carrier n sd I) = delta_mat_redtriple n sd 1 I"
| "get_redtriple (Arc_mat_carrier n I) = arctic_mat_redtriple n I"
| "get_redtriple (Arc_rat_mat_carrier n I) = arctic_delta_mat_redtriple n I"
| "get_redtriple (RPO prec\<tau> pi) = af_redtriple pi (create_RPO_redtriple (\<lambda> pr. (prec_repr_to_pr pr, prec_repr_to_status pr)) prec\<tau>)"
| "get_redtriple (KBO precw pi) = af_redtriple pi (create_KBO_redtriple (filter_prec_weight_repr (afs_list_to_afs pi)) precw)"

lemma get_redtriple: "generic_redtriple_impl get_redtriple"
proof
  fix rp :: "('f :: {key,show})redtriple_impl" and s_list ns_list nst_list:: "('f,'v::show)rules"
  let ?rp = "get_redtriple rp :: ('f,'v)redtriple"
  let ?all = "s_list @ ns_list @ nst_list"
  assume v: "isOK(redtriple.valid ?rp)" and S: "isOK(check_allm (redtriple.s ?rp) s_list)" and NS: "isOK(check_allm (redtriple.ns ?rp) ns_list)"
    and NST: "isOK(check_allm (redtriple.nst ?rp) nst_list)"
  note ass = v S NS NST
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST (redtriple.af ?rp) (redtriple.cpx ?rp) \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST \<and> (isOK(redtriple.mono ?rp ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST (redtriple.af ?rp) \<and> ctxt_closed S)"
  proof (cases rp)
    case (Int_carrier I)
    hence id: "?rp = (int_redtriple :: ('f,int) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" by simp
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF int_redtriple],
        insert ass id, auto)
  next
    case (Int_nl_carrier I)
    hence id: "?rp = (create_nlpoly_redtriple succeed (1 :: int) (op >) True True :: ('f,int)poly_inter_list \<Rightarrow> ('f,'v)redtriple) I" (is "_ = ?I I") by simp
    interpret mono_matrix_carrier "op >" 1 nat int_mono by (rule int_complexity)
    have "cpx_poly_order_carrier (1 :: int) (op >) True True nat" ..
    from poly_order_carrier_to_generic_redtriple_impl[OF this]
    have rp: "generic_redtriple_impl ?I" .
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]]) 
  next
    case (Rat_carrier I)
    hence id: "?rp = (delta_redtriple 1 :: ('f,rat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_redtriple],
        insert ass id, auto) 
  next
    case (Mini_alg_carrier I)
    hence id: "?rp = (delta_redtriple 1 :: ('f,mini_alg_number) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_redtriple],
        insert ass id, auto)
  next
    case (Rat_nl_carrier d I)
    hence id: "?rp = (delta_nl_redtriple d :: ('f,rat)poly_inter_list \<Rightarrow> ('f,'v)redtriple) I" by simp
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_nl_redtriple],
          insert ass id, auto)
  next
    case (Mini_alg_nl_carrier d I)
    hence id: "?rp = (delta_nl_redtriple d :: ('f,mini_alg_number)poly_inter_list \<Rightarrow> ('f,'v)redtriple) I" by simp
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_nl_redtriple],
          insert ass id, auto)
  next
    case (Arctic_carrier I)
    hence id: "?rp = (arctic_redtriple :: ('f,arctic) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF arctic_redtriple],
        insert ass id, auto)
  next
    case (Arctic_rat_carrier I)
    hence id: "?rp = (arctic_delta_redtriple :: ('f,rat arctic_delta) lpoly_interL \<Rightarrow> ('f,'v)redtriple) I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF arctic_delta_redtriple],
        insert ass id, auto)
  next
    case (Int_mat_carrier n sd I)
    hence id: "?rp = (int_mat_redtriple :: nat \<Rightarrow> nat \<Rightarrow> ('f,int mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n sd I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF int_mat_redtriple],
        insert ass id, auto)
  next
    case (Rat_mat_carrier n sd I)
    hence id: "?rp = (delta_mat_redtriple :: nat \<Rightarrow> nat \<Rightarrow> rat \<Rightarrow> ('f,rat mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n sd 1 I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_mat_redtriple],
        insert ass id, auto)
  next
    case (Mini_alg_mat_carrier n sd I)
    hence id: "?rp = (delta_mat_redtriple :: nat \<Rightarrow> nat \<Rightarrow> mini_alg_number \<Rightarrow> ('f,mini_alg_number mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n sd 1 I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF delta_mat_redtriple],
        insert ass id, auto)
  next
    case (Arc_mat_carrier n I)
    hence id: "?rp = (arctic_mat_redtriple :: nat \<Rightarrow> ('f,arctic mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF arctic_mat_redtriple],
        insert ass id, auto)
  next
    case (Arc_rat_mat_carrier n I)
    hence id: "?rp = (arctic_delta_mat_redtriple :: nat \<Rightarrow> ('f,rat arctic_delta mat) lpoly_interL \<Rightarrow> ('f,'v)redtriple) n I" by simp    
    show ?thesis unfolding id
      by (rule generic_redtriple_impl.generate_redtriple[OF arctic_delta_mat_redtriple],
        insert ass id, auto)
  next
    case (RPO prec\<tau> pi)
    let ?one = "(\<lambda> I. create_RPO_redtriple (\<lambda>pr. (prec_repr_to_pr pr, prec_repr_to_status pr)) I) :: 'f status_prec_repr \<Rightarrow> ('f filtered,'v)redtriple"
    have rp: "generic_redtriple_impl ?one" ..
    let ?two = "\<lambda> I. af_redtriple pi (?one I)"
    from af_redtriple[OF rp] 
    have rp: "generic_redtriple_impl ?two" .
    from RPO have id: "?rp = ?two prec\<tau>" by simp
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?two, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
  next
    case (KBO precw pi)
    let ?one = "(\<lambda> I. create_KBO_redtriple (filter_prec_weight_repr (afs_list_to_afs pi)) I) :: 'f prec_weight_repr \<Rightarrow> ('f filtered,'v)redtriple"
    have rp: "generic_redtriple_impl ?one" ..
    let ?two = "\<lambda> I. af_redtriple pi (?one I)"
    from af_redtriple[OF rp] 
    have rp: "generic_redtriple_impl ?two" .
    from KBO have id: "?rp = ?two precw" by simp
    show ?thesis by(simp only: id, rule generic_redtriple_impl.generate_redtriple[of ?two, OF rp  v[simplified id] S[simplified id] NS[simplified id] NST[simplified id]])
  qed 
qed

fun get_non_inf_order :: "('f :: {key,show}) redtriple_impl \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f,'v :: show)non_inf_order"
where 
  "get_non_inf_order (Int_nl_carrier I) = create_nlpoly_non_inf_order succeed (1 :: int) (op >) True True sqrt_int I"
| "get_non_inf_order (Rat_nl_carrier d I) = delta_non_inf_order d sqrt_rat I"
| "get_non_inf_order (Mini_alg_nl_carrier d I) = delta_non_inf_order d sqrt_mini_alg I"
| "get_non_inf_order _ = faulty_non_inf_order ''only integers, rationals and mini-alg-numbers supported for non-inf orders''"

lemma get_non_inf_order: "generic_non_inf_order_impl get_non_inf_order"
proof
  fix rp :: "('f :: {key,show})redtriple_impl" and F and ns_list :: "('f,'v::show)rules" and cc
  let ?rp = "get_non_inf_order rp F :: ('f,'v)non_inf_order"
  assume v: "isOK(non_inf_order.valid ?rp)" and cc: "isOK(check_allm (non_inf_order.cc ?rp) cc)" and NS: "isOK(check_allm (non_inf_order.ns ?rp) ns_list)"
  show "\<exists> S NS. non_inf_order S NS (set F) (non_inf_order.af ?rp) \<and> set ns_list \<subseteq> NS \<and> Ball (set cc) (cc_satisfied (set F) S NS)"
  proof (cases rp)
    case (Int_nl_carrier I)
    hence id: "?rp = (create_nlpoly_non_inf_order succeed (1 :: int) (op >) True True sqrt_int :: ('f,int)poly_inter_list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f,'v)non_inf_order) I F" (is "_ = ?I I F") by simp
    have "non_inf_poly_order_carrier (1 :: int) (op >) True True" 
      by (unfold_locales, insert non_inf_int_gt, auto simp: mult_right_mono_neg)
    from non_inf_poly_order_carrier_to_generic_non_inf_order[OF this]
    have rp: "generic_non_inf_order_impl ?I" .
    show ?thesis by (simp only: id, rule generic_non_inf_order_impl.generate_non_inf_order[of ?I, OF rp  v[simplified id] NS[simplified id] cc[simplified id]]) 
  next
    case (Rat_nl_carrier d I)
    hence id: "?rp = (delta_non_inf_order d sqrt_rat :: ('f,rat)poly_inter_list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f,'v)non_inf_order) I F" by simp
    show ?thesis unfolding id
      by (rule generic_non_inf_order_impl.generate_non_inf_order[OF delta_non_inf_order], insert NS v cc id, auto) 
  next
    case (Mini_alg_nl_carrier d I)
    hence id: "?rp = (delta_non_inf_order d sqrt_mini_alg :: ('f,mini_alg_number)poly_inter_list \<Rightarrow> ('f \<times> nat) list \<Rightarrow> ('f,'v)non_inf_order) I F" by simp
    show ?thesis unfolding id
      by (rule generic_non_inf_order_impl.generate_non_inf_order[OF delta_non_inf_order], insert NS v cc id, auto) 
  qed (insert v, auto simp: faulty_non_inf_order_def)
qed

datatype 'f root_redtriple_impl = SCNP list_order_type "'f scnp_af" "'f redtriple_impl"

fun get_root_redtriple :: "('f :: {key,show}) root_redtriple_impl \<Rightarrow> ('f,'v :: show)root_redtriple"
where 
  "get_root_redtriple (SCNP type af rp) = generate_scnp_rp (list_ext (scnp_arity af) type) (list_ext_name type) af get_redtriple rp"

lemma get_root_redtriple: "generic_root_redtriple_impl get_root_redtriple"
proof
  fix rrp :: "('f :: {key,show}) root_redtriple_impl" and s_list ns_list nst_list :: "('f, 'v :: show)rule list"
  let ?r = "get_root_redtriple rrp :: ('f,'v)root_redtriple"  
  assume ass: "isOK(root_redtriple.valid ?r)" 
         "isOK(check_allm (root_redtriple.s ?r) s_list)"
         "isOK(check_allm (root_redtriple.ns ?r) ns_list)" 
         "isOK(check_allm (root_redtriple.nst ?r) nst_list)" 
  show "\<exists> s ns nst. ce_af_root_redtriple_order s ns nst (root_redtriple.af ?r) (root_redtriple.aft ?r) \<and> set s_list \<subseteq> s \<and> set ns_list \<subseteq> ns \<and> set nst_list \<subseteq> nst"
  proof (cases rrp)
    case (SCNP t af rp)
    show ?thesis unfolding SCNP get_root_redtriple.simps
      by (rule generic_root_redtriple_impl.generate_root_redtriple
      [OF generate_scnp_rp[OF get_redtriple list_ext], OF ass[unfolded SCNP get_root_redtriple.simps]])
  qed
qed 

end
