(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Tree_Automata_Impl
imports
  Trs_Impl
  Tree_Automata
  Linear_Orders
  "$AFP/Transitive-Closure/RBT_Map_Set_Extension"
  "$AFP/Transitive-Closure/Transitive_Closure_RBT_Impl"
begin


datatype ('q, 'f) ta_rule_impl = TA_rule_impl 'f "'q list" 'q "'q rs" 

fun ta_rule_conv :: "('q, 'f) ta_rule_impl \<Rightarrow> ('q, 'f) ta_rule" where
  "ta_rule_conv (TA_rule_impl f qs q qs') = (TA_rule f qs q)"

fun conv_ta_rule :: "('q \<Rightarrow> 'q rs) \<Rightarrow> ('q, 'f) ta_rule \<Rightarrow> ('q, 'f) ta_rule_impl" where
  "conv_ta_rule eps (TA_rule f qs q) = (TA_rule_impl f qs q (eps q))"

record ('q, 'f) ta_impl =
  ta_final_impl :: "'q rs"
  ta_rules_impl :: "('f \<times> nat, ('q,'f) ta_rule_impl list) rm"
  ta_r_lhs_states_impl :: "'q list"
  ta_rhs_states_set :: "'q rs"
  ta_eps_impl :: "('q \<times> 'q) list"
  ta_epss_impl :: "'q \<Rightarrow> 'q rs"
  ta_epsrs_impl :: "'q \<Rightarrow> 'q rs"


definition ta_of :: "('q :: linorder, 'f :: linorder) ta_impl \<Rightarrow> ('q, 'f) ta" where
  "ta_of TA = \<lparr>
    ta_final = set (rs_to_list (ta_final_impl TA)), 
    ta_rules = \<Union>{ta_rule_conv ` (set (rm_set_lookup (ta_rules_impl TA) fn)) | fn. True},
    ta_eps = set (ta_eps_impl TA) \<rparr>"

fun r_lhs_states_impl where "r_lhs_states_impl (TA_rule_impl f qs q qs') = qs"
fun r_rhs_impl where "r_rhs_impl (TA_rule_impl f qs q qs') = q"
fun rqss_impl where "rqss_impl (TA_rule_impl f qs q qs') = qs'"
fun r_sym_impl where "r_sym_impl (TA_rule_impl f qs q qs') = (f,length qs)"

fun ta_res_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'q)term \<Rightarrow> 'q rs"
where "ta_res_impl TA eps (Var q) = eps q"
  | "ta_res_impl TA eps (Fun f ts) = (
        let rec = map (ta_res_impl TA eps) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = r_lhs_states_impl rule in Ball (set [0 ..< n]) (\<lambda> i. rs_memb (qs ! i) (rec ! i))) rules;
            qs = map (\<lambda> rule. rqss_impl rule) arules   
           in rs_Union qs)"



locale ta_inv = rm_set "ta_rules_impl TA" r_sym_impl "conv_ta_rule (ta_epss_impl TA) ` ta_rules (ta_of TA)" for
  TA :: "('q :: linorder,'f :: linorder)ta_impl" +
  fixes rel :: "'q rs \<Rightarrow> 'q rs \<Rightarrow> 'q option" and rell :: "'q rel" 
  assumes ta_rhs_states_set: "rs_\<alpha> (ta_rhs_states_set TA) = ta_rhs_states (ta_of TA)"
  and     ta_r_lhs_states_impl: "set (ta_r_lhs_states_impl TA) = ta_rhs_states (ta_of TA)"
  and     ta_epss: "\<And> q. rs_\<alpha> (ta_epss_impl TA q) = {q'. (q,q') \<in> (ta_eps (ta_of TA))^*}"
  and     ta_epsrs: "\<And> q. rs_\<alpha> (ta_epsrs_impl TA q) = {q'. (q',q) \<in> (ta_eps (ta_of TA))^*}"
  and     ta_eps: "ta_eps (ta_of TA) = set (ta_eps_impl TA)"
  and     rel: "\<And> ql qr. rel ql qr = None \<Longrightarrow> rs_\<alpha> ql \<subseteq> rell^-1 `` (rs_\<alpha> qr)"
  and     no_eps_or_rel_Id: "ta_eps (ta_of TA) = {} \<or> rell = Id"
  and     coherent: "state_coherent (ta_of TA) rell"
begin
abbreviation ta_rm_set_lookup where "ta_rm_set_lookup \<equiv> rm_set_lookup (ta_rules_impl TA)"
abbreviation TA' :: "('q,'f)ta" where "TA' \<equiv> ta_of TA"

lemma rm_set_lookup1: assumes "rule \<in> set (ta_rm_set_lookup (f,n))"
  shows "\<exists> qs q qs'. rule = (TA_rule_impl f qs q qs') \<and> 
    n = length qs \<and> rs_\<alpha> qs' = {q'. (q,q') \<in> (ta_eps TA')^*} \<and> (f qs \<rightarrow> q) \<in> ta_rules TA'"
proof -
  from assms have "r_sym_impl rule = (f,n)" and mem: "rule \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
    using assms unfolding rm_set_lookup by auto
  then obtain qs q qs' where n: "n = length qs" and rule: "rule = TA_rule_impl f qs q qs'" by (cases rule, auto)
  from mem obtain rrule where id: "rule = conv_ta_rule (ta_epss_impl TA) rrule" and rrule: "rrule \<in> ta_rules TA'" by auto
  from id[unfolded rule] have idd: "rrule = (f qs \<rightarrow> q)" by (cases rrule, auto) 
  from id[unfolded rule idd] have "rs_\<alpha> qs' = rs_\<alpha> (ta_epss_impl TA q)" by auto
  from this[unfolded ta_epss] rrule
  show ?thesis unfolding idd rule n
    by auto
qed

lemma rm_set_lookup2: assumes "rule \<in> conv_ta_rule (ta_epss_impl TA) ` (ta_rules TA')"
  shows "rule \<in> set (ta_rm_set_lookup (r_sym_impl rule))"
  using assms unfolding rm_set_lookup
  by (cases rule, auto)

lemma rm_set_lookup3: assumes "rule \<in> conv_ta_rule (ta_epss_impl TA) ` (ta_rules TA')"
  shows "\<exists> rules. Some rules = rm_\<alpha> (ta_rules_impl TA) (r_sym_impl rule) \<and> rule \<in> set rules"
  using rm_set_lookup2[OF assms] by (auto simp: rm_set_lookup_def)


lemma ta_res_impl: 
  shows "rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) t) = ta_res TA' t"
proof (induct t)
  case (Var x) thus ?case by (simp add: ta_epss)
next
  case (Fun f ts)
  show ?case
  proof (rule set_eqI)
    fix q
    show "(q \<in> rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (Fun f ts))) = (q \<in> ta_res (ta_of TA) (Fun f ts))"
      (is "(q \<in> ?L) = (q \<in> ?S)")
    proof
      assume "q \<in> ?L"
      then obtain qs' rule where mem: "rule \<in> set (ta_rm_set_lookup (f,length ts))"
        and q: "rqss_impl rule = qs'"
        and q': "q \<in> rs_\<alpha> qs'"
        and args: "\<forall> i \<in> {0 ..< length ts}.
             rs_memb (r_lhs_states_impl rule ! i) (map (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA)) ts ! i)"
        by (auto simp: Let_def rs_correct)
      from rm_set_lookup1[OF mem] 
      obtain qs qs'' q' where id: "rule = TA_rule_impl f qs q' qs''" and
        len: "length ts = length qs" and 
        qs'': "rs_\<alpha> qs'' = {q. (q',q) \<in> (ta_eps TA')^*}" and
        rule: "(f qs \<rightarrow> q') \<in> ta_rules TA'" by auto
      from q id have qs': "qs' = qs''" by auto
      from q'[unfolded qs' qs''] have qq': "(q',q) \<in> (ta_eps TA')^*" by simp
      show "q \<in> ?S"
      proof (unfold ta_res.simps, rule, rule exI[of _ q], rule exI[of _ q'], rule exI[of _ qs], intro conjI, simp, simp add: rule, rule qq', simp add: len, intro allI impI)
        fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by auto
        from args[unfolded id, simplified] i len
        have "rs_memb (qs ! i) (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (ts ! i))" by auto
        with Fun[OF mem] i 
        show "qs ! i \<in> map (ta_res TA') ts ! i" by (simp add: rs_correct)
      qed
    next
      assume "q \<in> ?S"
      then obtain q' qs where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA'"
        and len: "length qs = length ts"
        and args: "\<forall> i < length ts. qs ! i \<in> map (ta_res TA') ts ! i"
        and q: "(q',q) \<in> (ta_eps TA')^*"
        by auto
      let ?rule = "TA_rule_impl f qs q' (ta_epss_impl TA q')"
      have rrule: "?rule = conv_ta_rule (ta_epss_impl TA) (f qs \<rightarrow> q')" by auto
      have rrule: "?rule \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
        unfolding rrule using rule by blast
      have rule: "?rule \<in> set (ta_rm_set_lookup (f,length qs))"
        using rm_set_lookup3[OF rrule]
        unfolding rm_set_lookup_def by auto
      {
                fix i
        assume i: "i < length ts"
        hence mem: "ts ! i \<in> set ts" by simp
        have "qs ! i \<in> rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (ts ! i))"
          unfolding Fun[OF mem] using args i by auto
      } note args = this
      show "q \<in> ?L"
        by (simp add: Let_def rs_correct, intro exI conjI, rule rule[unfolded len], 
          insert args q, simp add: len, simp add: ta_epss)
    qed
  qed
qed
end



(* the following performs a brute force enumeration over
  all possible substitutions; this is a bottle-neck and an improved version is developed below:
  rule_state_compatible_eff_list *)
definition rule_state_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('q rs \<Rightarrow> 'q rs \<Rightarrow> 'q option)  \<Rightarrow> ('f,'v)rule \<times> 'v list \<Rightarrow> (('f,'q)rule \<times> 'q) check"
where "rule_state_compatible_list TA rel \<equiv> let rm = ta_rules_impl TA; eps = ta_epss_impl TA in (\<lambda> ((l,r),xs). 
          (check_allm (\<lambda> vec. let \<tau> = fun_of vec;
                               l\<tau> = map_vars \<tau> l;
                               r\<tau> = map_vars \<tau> r;
                               rhs = ta_res_impl rm eps r\<tau>;
                               lhs = ta_res_impl rm eps l\<tau>
                            in  (case rel lhs rhs of None \<Rightarrow> succeed | Some q \<Rightarrow> error ((l\<tau>,r\<tau>),q)))
                        (enum_vectors (ta_r_lhs_states_impl TA) xs)))"

context ta_inv
begin

lemma rule_state_compatible_list: assumes 
     wf: "vars_term r \<subseteq> vars_term l"
  and comp: "isOK(rule_state_compatible_list TA rel ((l,r),vars_term_list l))"
  shows "rule_state_compatible TA' rell (l,r)"
proof (cases "ta_rhs_states TA' = {}")
  case False
  show ?thesis
    unfolding rule_state_compatible_def 
  proof(clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rhs_states TA'"
    let ?Q = "ta_r_lhs_states_impl TA"
    from False have "?Q \<noteq> []" using ta_r_lhs_states_impl by auto
    from enum_vectors_complete[OF this, of "vars_term_list l"]
    obtain vec where vec: "vec \<in> set (enum_vectors ?Q (vars_term_list l))" 
      and \<tau>2: "\<And> x c. x \<in> set (vars_term_list l) \<Longrightarrow> c \<in> set ?Q \<Longrightarrow>  \<tau> x = c \<longrightarrow> fun_of vec x = c" by blast
    let ?\<tau> = "fun_of vec"
    {
      fix x
      assume x: "x \<in> vars_term l"
      with \<tau> have "\<tau> x \<in> ta_rhs_states TA'" by auto    
      from \<tau>2[unfolded vars_term_list_sound ta_r_lhs_states_impl, OF x this] have "\<tau> x = ?\<tau> x"
        by simp
    } note \<tau>2 = this
    have l: "map_vars \<tau> l = map_vars ?\<tau> l"
      by (rule map_vars_vars_term[OF \<tau>2])
    have r: "map_vars \<tau> r = map_vars ?\<tau> r"
      by (rule map_vars_vars_term, insert \<tau>2 wf, auto)
    let ?R = "ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA)"    
    let ?check = "rel (?R (map_vars (fun_of vec) l)) (?R (map_vars (fun_of vec) r))"
    from comp[unfolded rule_state_compatible_list_def Let_def, simplified, THEN bspec[OF _ vec]] 
    have "?check = None" by (cases ?check, auto)
    from rel[OF this, unfolded ta_res_impl] q
    show "q \<in> rell^-1 `` ta_res TA' (map_vars \<tau> r)" unfolding l r by auto
  qed
next
  case True
  show "rule_state_compatible TA' rell (l,r)"
    unfolding rule_state_compatible_def
  proof (clarify)
    fix \<tau> q
    assume q: "q \<in> ta_res TA' (map_vars \<tau> l)"
      and \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rhs_states TA'"    
    show "q \<in> rell^-1 `` ta_res TA' (map_vars \<tau> r)"
    proof (cases l)
      case (Fun f ls)
      from q[unfolded Fun] True
      show ?thesis unfolding ta_rhs_states_def by auto
    next
      case (Var x)
      from \<tau>[unfolded Var ta_rhs_states_def]  True
      show ?thesis unfolding ta_rhs_states_def by simp
    qed
  qed    
qed
end

(* TODO: integrate (here and in ta_match) an optimization of concat / fun_merge that deals with 
   non-linear terms. Currently conflicts might occur (x / a and x / b) and duplicates may occur (x / a 
   and x / a). This is only important for efficiency, it will not effect power *)
fun ta_match_pre_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v \<times> 'q)list set"
  where "ta_match_pre_impl TA eps (Var x) q = { [(x,q')] | q'. q' \<in> rs_\<alpha> (eps q)}"
   | "ta_match_pre_impl TA eps (Fun f ts) q = { concat \<sigma>s | \<sigma>s qs' qs q'. 
          (TA_rule_impl f qs q' qs') \<in> set (rm_set_lookup TA (f,length ts)) \<and> 
          q' \<in> rs_\<alpha> (eps q) \<and> 
          length \<sigma>s = length ts \<and>
          (\<forall> i < length ts. \<sigma>s ! i \<in> ta_match_pre_impl TA eps (ts ! i) (qs ! i)) }"

context ta_inv
begin
lemma ta_match_pre_impl': "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q. (set (map fst \<tau>) = vars_term t \<and> (\<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x))"
proof (induct t arbitrary: q \<sigma>)
  case (Var x q)
  thus ?case  unfolding fun_of_def using ta_epsrs
    by auto
next
  case (Fun f ts q \<sigma>)
  let ?I = "ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA)"
  from Fun(2) obtain \<sigma>s qs q' where rule: "(f qs \<rightarrow> q') \<in> ta_rules TA'"
    and q'q: "(q',q) \<in> (ta_eps TA')^*"
    and \<sigma>: "\<sigma> = fun_merge \<sigma>s (map vars_term ts)"
    and len1: "length qs = length ts"
    and len2: "length \<sigma>s = length ts" 
    and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> ta_match TA' (ts ! i) (qs ! i)" by auto
  let ?prop = "\<lambda> \<tau> i. \<tau> \<in> ?I (ts ! i) (qs ! i) \<and> set (map fst \<tau>) = vars_term (ts ! i) \<and> (\<forall> x \<in> vars_term (ts ! i). fun_of \<tau> x = (\<sigma>s ! i) x)"
  {
    fix i      
    assume i: "i < length ts"
    let ?t = "ts ! i"
    from i have t: "?t \<in> set ts" by auto
    from Fun(1)[OF t rec[OF i]] have "\<exists> \<tau>. ?prop \<tau> i" by auto
  }
  hence "\<forall> i. \<exists> \<tau>. i < length ts \<longrightarrow> ?prop \<tau> i" by auto
  from choice[OF this] obtain \<tau>s where \<tau>s: "\<And> i. i < length ts \<Longrightarrow> ?prop (\<tau>s i) i" by auto
  obtain \<tau> where \<tau>: "\<tau> = concat (map (\<lambda> i. \<tau>s i) [0 ..< length ts])" by auto
  have vars_term: "set (map fst \<tau>) = vars_term (Fun f ts)" (is "?L = ?R")
  proof (unfold set_eq_iff, rule)
    fix x
    {
      assume "x \<in> ?L"
      from this[unfolded \<tau>] obtain i where i: "i < length ts" and x: "x \<in> set (map fst (\<tau>s i))" by force
      from \<tau>s[OF i] x have x: "x \<in> vars_term (ts ! i)" by auto
      hence "x \<in> ?R" using i 
        by (auto simp: set_conv_nth[of ts])
    }
    moreover
    {
      assume "x \<in> ?R"
      then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
        by (auto simp: set_conv_nth[of ts])
      from \<tau>s[OF i] x have x: "x \<in> set (map fst (\<tau>s i))" by force
      with i have "x \<in> ?L" unfolding \<tau> by force
    }
    ultimately show "(x \<in> ?L) = (x \<in> ?R)" by auto
  qed
  show ?case
  proof(intro bexI[of _ \<tau>] conjI, rule vars_term, intro ballI)
    show "\<tau> \<in> ?I (Fun f ts) q"
    proof (unfold ta_match_pre_impl.simps, rule, unfold ta_epsrs, intro exI conjI allI impI, rule \<tau>)
      fix i
      assume i: "i < length ts"
      show "map \<tau>s [0 ..< length ts] ! i \<in> ?I (ts ! i) (qs ! i)"
        using \<tau>s[OF i, THEN conjunct1] i by auto
    next
      show "q' \<in> {q'. (q',q) \<in> (ta_eps TA')^*}" using q'q by simp
    next
      show "length (map \<tau>s [0..< length ts]) = length ts" by simp
    next
      let ?e = "ta_epss_impl TA q'"
      show "TA_rule_impl f qs q' ?e \<in> set (ta_rm_set_lookup (f, length ts))"
      proof (unfold rm_set_lookup, rule, intro conjI)
        show "r_sym_impl (TA_rule_impl f qs q' ?e) = (f, length ts)"
          by (simp add: len1)
      next
        show "TA_rule_impl f qs q' ?e \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'"
          by (rule, insert rule, auto)
      qed
    qed
  next
    fix x
    assume "x \<in> ?R"
    then obtain i where i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
      by (auto simp: set_conv_nth[of ts])
    show "fun_of \<tau> x = \<sigma> x" unfolding \<sigma> \<tau>
    proof (rule fun_of_concat_merge)
      fix i
      assume i: "i < length ts" and x: "x \<in> vars_term (ts ! i)"
      show "fun_of (\<tau>s i) x = (\<sigma>s ! i) x"
        using \<tau>s[OF i] x i by auto
    next
      show "length ts = length \<sigma>s" using len1 len2 by auto
    next
      fix i
      assume i: "i < length ts"
      show "(map_of (\<tau>s i) x \<noteq> None) = (x \<in> vars_term (ts ! i))"
      proof -
        have "(x \<in> vars_term (ts ! i)) = (x \<in> set (map fst (\<tau>s i)))" (is "?l = ?m")
          using \<tau>s[OF i]  by simp
        also have "... = (map_of (\<tau>s i) x \<noteq> None)" (is "_ = ?r")
        proof
          assume ?m
          then obtain y where "(x,y) \<in> set (\<tau>s i)" by auto
          with map_of_eq_None_iff[of "\<tau>s i" x] show ?r by force
        next
          assume ?r
          then obtain y where "map_of (\<tau>s i) x = Some y" by (cases "map_of (\<tau>s i) x", auto)
          from map_of_SomeD[OF this] show ?m by force
        qed
        finally show "?r = ?l" by simp
      qed
    qed (insert i x, auto)
  qed
qed

lemma ta_match_pre_impl: "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> ta_match_pre_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q. \<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x"
  using ta_match_pre_impl' by blast
end

fun ta_match_impl :: "('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('q \<Rightarrow> 'q rs) \<Rightarrow> ('f,'v)term \<Rightarrow> 'q \<Rightarrow> ('v :: linorder \<times> 'q)list rs"
  where "ta_match_impl TA eps (Var x) q = list_to_rs (map (\<lambda>q'. [(x,q')]) (rs_to_list (eps q)))" 
   | "ta_match_impl TA eps (Fun f ts) q = (let 
          n = length ts;
          rules = rm_set_lookup TA (f, n);
          ep = eps q;
          f = (\<lambda> rule. list_to_rs (case rule of TA_rule_impl _ qs q' qs' \<Rightarrow>
                 (if rs_memb q' ep then 
                   (let rec = map (\<lambda> (tsi,qsi). rs_to_list (ta_match_impl TA eps tsi qsi)) (zip ts qs)
                     in map concat (concat_lists rec))
                 else [])))
        in rs_Union (map f rules))" 

context ta_inv
begin
lemma ta_match_impl_conv_pre: "rs_\<alpha> (ta_match_impl (ta_rules_impl TA) eps t q) = ta_match_pre_impl (ta_rules_impl TA) eps t q"
proof (induct t arbitrary: q)
  case (Var x)
  thus ?case by (auto simp: rs_correct)
next
  case (Fun f ts q)
  let ?TA = "ta_rules_impl TA"
  show ?case (is "?I = ?P")
  proof -
    {
      fix \<sigma>
      assume "\<sigma> \<in> ?P"
      from this
      obtain \<sigma>s qs qs' q' where \<sigma>: "\<sigma> = concat \<sigma>s"
        and rule: "TA_rule_impl f qs q' qs' \<in> set (ta_rm_set_lookup (f, length ts))"
        and q'q: "q' \<in> rs_\<alpha> (eps q)"
        and len1: "length \<sigma>s = length ts"
        and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> ta_match_pre_impl ?TA eps (ts ! i) (qs ! i)" by auto
      from rm_set_lookup1[OF rule] have len2: "length qs = length ts" by auto
      from q'q have q'q: "rs_memb q' (eps q)" by (simp add: rs_correct)
      have " \<exists>a\<in>set (ta_rm_set_lookup (f, length ts)).
       \<exists>uu q' qs.
          (\<exists> f. TA_rule_impl f qs q' uu = a) \<and>
          rs_memb q' (eps q) \<and>
          concat \<sigma>s
          \<in> concat `
            {as. length as = min (length ts) (length qs) \<and>
                 (\<forall>i. i < length ts \<and> i < length qs \<longrightarrow>
                      as ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i)))} "
      proof (rule bexI[OF _ rule], intro exI conjI, rule refl, rule q'q, rule, rule refl, rule,
        unfold len1 len2, intro conjI allI impI allI, simp)
        fix i
        assume "i < length ts \<and> i < length ts" 
        hence i: "i < length ts" by auto
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] rec[OF i] show "\<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i))" by auto
      qed
      hence "\<sigma> \<in> ?I" unfolding ta_match_impl.simps Let_def set_concat rs_correct \<sigma>
        by (simp add: rs_correct)
    }
    moreover
    {
      fix \<sigma>
      assume "\<sigma> \<in> ?I"
      from this
      obtain rule :: "('q,'f)ta_rule_impl" and g and \<sigma>s :: "'q \<Rightarrow> 'q" and qs' qs q' where
         rule: "TA_rule_impl g qs q' qs' \<in> set (ta_rm_set_lookup (f, length ts))"
        and q': "rs_memb q' (eps q)"
        and \<sigma>: "\<sigma> \<in> concat ` {\<sigma>s. length \<sigma>s = min (length ts) (length qs) \<and> (\<forall> i. i < length ts \<and> i < length qs \<longrightarrow> \<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i)))}" by (simp add: rs_correct) blast+
      from rm_set_lookup1[OF rule] have gf: "g = f" and len2: "length qs = length ts" by auto
      from \<sigma> obtain \<sigma>s where \<sigma>: "\<sigma> = concat \<sigma>s" and len1: "length \<sigma>s = length ts" and rec: "\<And> i. i < length ts \<Longrightarrow> \<sigma>s ! i \<in> rs_\<alpha> (ta_match_impl ?TA eps (ts ! i) (qs ! i))" unfolding len2 by auto
      have "\<sigma> \<in> ?P"
      proof (simp, intro exI conjI, rule \<sigma>, rule rule[unfolded gf], rule q'[simplified rs_correct], rule len1,
        intro allI impI)
        fix i
        assume i: "i < length ts" 
        hence "ts ! i \<in> set ts" by auto
        from Fun[OF this] rec[OF i] show "\<sigma>s ! i \<in> ta_match_pre_impl ?TA eps (ts ! i) (qs ! i)" by auto
      qed
    }
    ultimately show ?thesis by blast
  qed
qed        

lemma ta_match_impl: "\<sigma> \<in> ta_match TA' t q \<Longrightarrow> \<exists> \<tau> \<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) t q). \<forall> x \<in> vars_term t. fun_of \<tau> x = \<sigma> x"
  using ta_match_pre_impl unfolding ta_match_impl_conv_pre .
end

declare ta_match_impl.simps[simp del]

fun rule_state_compatible_eff_list :: "('q :: linorder,'f :: linorder)ta_impl 
  \<Rightarrow> ('q rs \<Rightarrow> 'q rs \<Rightarrow> 'q option)
  \<Rightarrow> ('f,'v :: linorder)rule 
  \<Rightarrow> (('f,'q)rule \<times> 'q) check"
where "rule_state_compatible_eff_list TA rel (Fun f ls,r) =  
           (let rm = ta_rules_impl TA; 
                eps = ta_epss_impl TA;
                eps' = ta_epsrs_impl TA;
                n = length ls;
                ta_res' = ta_res_impl rm eps
             in check_allm (\<lambda> rule. case rule of  TA_rule_impl f qs q qs' \<Rightarrow>
                   check_allm (\<lambda> \<sigma>. 
                       let \<sigma>' = fun_of \<sigma>;
                           l\<sigma> = map_vars \<sigma>' (Fun f ls);
                           r\<sigma> = map_vars \<sigma>' r;
                           qsl = ta_res' l\<sigma>;
                           qsr = ta_res' r\<sigma>
                         in (case rel qsl qsr of None \<Rightarrow> succeed | Some q' \<Rightarrow> error ((l\<sigma>,r\<sigma>),q'))
                   ) (rs_to_list (ta_match_impl rm eps' (Fun f ls) q)))
                 (rm_set_lookup rm (f,n)))" |
    "rule_state_compatible_eff_list TA rel (Var x,r) = 
           (let rm = ta_rules_impl TA; 
                eps = ta_epss_impl TA;
                ta_res' = ta_res_impl rm eps
             in 
                check_allm (\<lambda> q. 
                  (let r\<sigma> = map_vars (\<lambda> _. q) r
                    in case rel (list_to_rs [q]) (ta_res' r\<sigma>) of None \<Rightarrow> succeed | Some q' \<Rightarrow> error ((Var q, r\<sigma>),q)))
                (ta_r_lhs_states_impl TA))"

lemma (in ta_inv) rule_state_compatible_eff_list: assumes 
    ok: "isOK(rule_state_compatible_eff_list TA rel (l,r))"
  and vars: "vars_term r \<subseteq> vars_term l"
  shows "rule_state_compatible TA' rell (l,r)"
proof (rule rule_state_compatible_eff[OF vars no_eps_or_rel_Id])
  show "rule_state_compatible_eff TA' rell (l, r)" 
  proof (cases l)
    case (Fun f ls)
    show ?thesis unfolding Fun rule_state_compatible_eff.simps
    proof (intro allI impI, clarify)
      fix \<sigma> qs q
      assume rule: "(f qs \<rightarrow> q) \<in> ta_rules TA'" and 
        len: "length qs = length ls" and 
        match: "\<sigma> \<in> ta_match TA' (Fun f ls) q" and
        q: "q \<in> ta_res TA' (map_vars \<sigma> (Fun f ls))"
      from ta_match_impl[OF match]
      obtain \<tau> where \<tau>1: "\<tau> \<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) l q)"
        and \<tau>2: "\<And> x. x \<in> vars_term l \<Longrightarrow> fun_of \<tau> x = \<sigma> x" unfolding Fun by blast
      let ?\<tau> = "fun_of \<tau>"
      have "map_vars \<sigma> (Fun f ls) = map_vars ?\<tau> l" unfolding Fun[symmetric]
        by (rule map_vars_vars_term, simp add: \<tau>2)
      from q[unfolded this] have q: "q \<in> ta_res TA' (map_vars ?\<tau> l)" by auto
      let ?e = "ta_epss_impl TA q"
      have rule: "TA_rule_impl f qs q ?e \<in> set (ta_rm_set_lookup (f,length ls))"
        unfolding rm_set_lookup
        by (rule, intro conjI, rule image_eqI[OF _ rule], simp_all add: len)
      let ?rl = "\<lambda> x. ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars (fun_of x) l)"
      let ?rr = "\<lambda> x. ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars (fun_of x) r)"
      from ok[unfolded Fun rule_state_compatible_eff_list.simps Let_def] rule
      have " \<And>x. x\<in> rs_\<alpha> (ta_match_impl (ta_rules_impl TA) (ta_epsrs_impl TA) l q) \<Longrightarrow>
        isOK (case rel (?rl x) (?rr x) of
           None \<Rightarrow> Inr () | Some q' \<Rightarrow> Inl ((map_vars (fun_of x) l, map_vars (fun_of x) r), q'))" 
        unfolding Fun by (auto simp: rs_correct)
      from this[OF \<tau>1[unfolded Fun[symmetric]]]
      have "rel (?rl \<tau>) (?rr \<tau>) = None"
        by (cases "rel (?rl \<tau>) (?rr \<tau>)", auto)
      from rel[OF this, unfolded  ta_res_impl] q
      have q: "q \<in> rell^-1 `` ta_res TA' (map_vars ?\<tau> r)" by auto
      have "map_vars \<sigma> r = map_vars ?\<tau> r"
        by (rule map_vars_vars_term, insert vars \<tau>2, auto)
      thus "\<exists> q''. (q,q'') \<in> rell \<and> q'' \<in> ta_res TA' (map_vars \<sigma> r)" using q by auto
    qed
  next
    case (Var x)
    show ?thesis
      unfolding Var rule_state_compatible_eff.simps
    proof (clarify)
      fix q 
      assume q: "q \<in> ta_rhs_states TA'"
      let ?check = "rel (list_to_rs [q]) (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars (\<lambda>_. q) r))"
      from ok[unfolded Var] q have "?check = None" by (cases ?check, auto simp: Let_def ta_res_impl ta_r_lhs_states_impl)
      from rel[OF this]
      show "\<exists> q''. (q,q'') \<in> rell \<and> q'' \<in> ta_res TA' (map_vars (\<lambda>_. q) r)"
        by (auto simp: Let_def ta_r_lhs_states_impl rs_correct ta_res_impl rel)
    qed
  qed
qed


fun ta_res_impl_all :: "'q rs \<Rightarrow> ('f \<times> nat,('q :: linorder,'f :: linorder)ta_rule_impl list)rm \<Rightarrow> ('f,'v)term \<Rightarrow> 'q rs"
where "ta_res_impl_all Q TA (Var _) = Q"
  | "ta_res_impl_all Q TA (Fun f ts) = (
        let rec = map (ta_res_impl_all Q TA) ts;
            n = length ts;
            rules = rm_set_lookup TA (f,n);
            arules = filter (\<lambda> rule. let qs = r_lhs_states_impl rule in Ball (set [0 ..< n]) (\<lambda> i. rs_memb (qs ! i) (rec ! i))) rules;
            qs = map (\<lambda> rule. rqss_impl rule) arules   
           in rs_Union qs)"

context ta_inv
begin
lemma ta_res_impl_all: assumes "\<tau> ` vars_term t \<subseteq> ta_rhs_states TA'"
  shows "rs_\<alpha> (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA) (map_vars \<tau> t)) 
    \<subseteq> rs_\<alpha> (ta_res_impl_all (ta_rhs_states_set TA) (ta_rules_impl TA) t)"
  using assms
proof (induct t)
  case (Var q)
  with assms  have q: "\<tau> q \<in> ta_rhs_states TA'" by auto
  { 
    fix q'
    assume q': "(\<tau> q, q') \<in> (ta_eps TA')^*"
    from q q' 
    have "q' \<in> ta_rhs_states TA'" unfolding ta_rhs_states_def
      by auto
  }
  thus ?case by (auto simp: ta_epss ta_rhs_states_set)
next
  case (Fun f ts)
  {
    fix rule :: "('q,'f)ta_rule_impl"
    assume 
      args: "\<forall> i \<in> {0..<length ts}. rs_memb (r_lhs_states_impl rule ! i) 
        (map (ta_res_impl (ta_rules_impl TA) (ta_epss_impl TA)) (map (map_vars \<tau>) ts) ! i)"
    have args: "\<forall> i \<in> {0..<length ts}. rs_memb (r_lhs_states_impl rule ! i)  
      (map (ta_res_impl_all (ta_rhs_states_set TA) (ta_rules_impl TA)) ts ! i)" (is "\<forall> i \<in> _. ?c i")
    proof -
      {
        fix i
        assume i: "i < length ts"
        from i have mem: "ts ! i \<in> set ts" by auto
        with Fun(2) have "\<tau> ` vars_term (ts ! i) \<subseteq> ta_rhs_states TA'" by auto
        from Fun(1)[OF mem this] args i
        have "?c i" by (auto simp: rs_correct)
      }
      thus ?thesis by auto
    qed
  }
  thus ?case by (auto simp: Let_def rs_correct)
qed
end

definition rule_state_compatible_heuristic :: "('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> ('f,'v :: linorder)term \<Rightarrow> bool" where
"rule_state_compatible_heuristic TA l \<equiv> rs_isEmpty (ta_res_impl_all (ta_rhs_states_set TA) (ta_rules_impl TA) l)"

context ta_inv
begin

lemma rule_state_compatible_heuristic_subteq: assumes compat: "rule_state_compatible_heuristic TA u"
    and subt: "l \<unrhd> u"
  shows "rule_state_compatible TA' rell (l,r)"
proof -
  {
    fix \<tau>
    assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rhs_states TA'"
    {
      fix x
      assume "x \<in> vars_term u"
      from subt supteq_Var[OF this] have "l \<unrhd> Var x" by (rule supteq_trans)
      from var_subt_imp_subt_in_vars[OF this] have "x \<in> vars_term l" .
    }
    with \<tau> have \<tau>u: "\<tau> ` vars_term u \<subseteq> ta_rhs_states TA'" by auto
    from assms[unfolded rule_state_compatible_heuristic_def]
    have empty: "rs_\<alpha> (ta_res_impl_all (ta_rhs_states_set TA) (ta_rules_impl TA) u) = {}" by (simp add: rs_correct)
    from ta_res_impl_all[OF \<tau>u] empty
    have empty: "ta_res TA' (map_vars \<tau> u) = {}" 
      unfolding ta_res_impl by auto
    from subt obtain C where l: "l = C\<langle>u\<rangle>" ..
    have "ta_res TA' (map_vars \<tau> l) = {}"
      unfolding l
    proof (induct C)
      case (More f bef C aft)
      { 
        fix qs
        have "\<And> g. \<exists> i < Suc (length bef + length aft). qs ! i \<notin> (map g bef @ ta_res TA' (map_vars \<tau> C\<langle>u\<rangle>) # map g aft) ! i"
          by (rule exI[of _ "length bef"], simp add: nth_append, insert More, auto)
      }
      thus ?case by (auto simp: o_def)
    qed (insert empty, auto)
  }
  thus ?thesis
    unfolding rule_state_compatible_def by auto
qed
end


definition state_compatible_list :: "('q :: linorder,'f :: linorder)ta_impl 
  \<Rightarrow> ('q rs \<Rightarrow> 'q rs \<Rightarrow> 'q option)
  \<Rightarrow> (('f,'v)rule \<times> 'v list)list \<Rightarrow> (('f,'v)rule \<times> ('f,'q)rule \<times> 'q) check"
where "state_compatible_list TA rel R \<equiv> let check = rule_state_compatible_list TA rel in
     check_allm (\<lambda> lr. check lr <+? (\<lambda> lq. (fst lr,lq))) R"

lemma (in ta_inv) state_compatible_list: assumes wf_trs: "\<And> l r. (l,r) \<in> fst ` set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and vars: "\<And> l r xs. ((l,r),xs) \<in> set R \<Longrightarrow> xs = vars_term_list l"
  and comp: "isOK(state_compatible_list TA rel R)"
  shows "state_compatible TA' rell (fst ` set R)" 
proof -
  {
    fix l r xs
    assume rule: "((l,r),xs) \<in> set R"
    from comp[unfolded state_compatible_list_def Let_def]
      rule have comp: "isOK(rule_state_compatible_list TA rel ((l,r),xs))"
      by auto
    from rule_state_compatible_list[OF wf_trs comp[unfolded vars[OF rule]]] rule
    have "rule_state_compatible TA' rell (l,r)" by force
  }
  thus ?thesis unfolding state_compatible_def 
    by force
qed

definition state_compatible_eff_list :: 
  "('q :: linorder,'f :: linorder)ta_impl 
  \<Rightarrow> ('q rs \<Rightarrow> 'q rs \<Rightarrow> 'q option)
  \<Rightarrow> ('f,'v :: linorder)rule list \<Rightarrow> (('f,'v)rule \<times> ('f,'q)rule \<times> 'q) check"
where "state_compatible_eff_list TA rel R \<equiv> let check = rule_state_compatible_eff_list TA rel in
     check_allm (\<lambda> lr. check lr <+? (\<lambda> lq. (lr,lq))) R"

lemma (in ta_inv) state_compatible_eff_list: assumes wf_trs: "\<And> l r. (l,r) \<in> set R \<Longrightarrow> vars_term r \<subseteq> vars_term l"
  and comp: "isOK(state_compatible_eff_list TA rel R)"
  shows "state_compatible TA' rell (set R)" 
proof -
  {
    fix l r 
    assume rule: "(l,r) \<in> set R"
    from comp[unfolded state_compatible_eff_list_def Let_def]
      rule have comp: "isOK(rule_state_compatible_eff_list TA rel (l,r))"
      by auto
    from rule_state_compatible_eff_list[OF comp wf_trs[OF rule]]
    have "rule_state_compatible TA' rell (l,r)" .
  }
  thus ?thesis unfolding state_compatible_def 
    by force
qed


definition ta_contains_impl :: "('f \<times> nat) list \<Rightarrow> ('q :: linorder,'f :: linorder)ta_impl \<Rightarrow> 'q list \<Rightarrow> ('q + ('f,'q)term)check"
  where "ta_contains_impl F TA qs \<equiv> 
       do { 
         let fin = ta_final_impl TA;
         let look = rm_set_lookup (ta_rules_impl TA);
         let eps = ta_epss_impl TA;
         check_allm (\<lambda> q. check (rs_memb q fin) q) qs
           <+? Inl;
         check_allm (\<lambda> (f,n). let rules = look (f,n) 
              in (check_all (\<lambda> qs'. Bex (set rules) (\<lambda> rule. r_lhs_states_impl rule = qs' 
                \<and> (let qq = rqss_impl rule in Bex (set qs) (\<lambda>q. rs_memb q qq))))
          (concat_lists (replicate n qs))) <+? (\<lambda>qs'. Fun f (map Var qs'))) F
            <+? Inr
          } "

context ta_inv
begin
(* I believe that also the other direction is correct, but currently this
   is not needed *)
lemma ta_contains_impl: 
  assumes contain: "isOK(ta_contains_impl F TA qs)"
  shows "ta_contains (set F) TA' (set qs)"
proof -
  note contain = contain[unfolded ta_contains_impl_def Let_def set_concat_lists, simplified, simplified rs_correct, simplified]
  have final: "set qs \<subseteq> ta_final TA'" using contain[THEN conjunct1]
    unfolding ta_of_def by (auto simp: rs_correct)
  show ?thesis unfolding ta_contains_def
  proof (rule conjI[OF final], intro allI impI)
    fix f qs'
    assume "(f,length qs') \<in> set F \<and> set qs' \<subseteq> set qs"
    hence f: "(f,length qs') \<in> set F" and i: "\<forall> i < length qs'. qs' ! i \<in> set qs" unfolding set_conv_nth[of qs'] by auto
    from contain[THEN conjunct2, THEN bspec[OF _ f], simplified, THEN spec[of _ qs']] i obtain rule q where q: "q \<in> set qs" and rule: "rule \<in> set (ta_rm_set_lookup (f, length qs'))" 
      and q2: "q \<in> rs_\<alpha> (rqss_impl rule)" and qs': "r_lhs_states_impl rule = qs'" by auto
    from rm_set_lookup1[OF rule] 
    obtain qsr1 qr qsr2 where rule: "rule = TA_rule_impl f qsr1 qr qsr2"
      and len: "length qs' = length qsr1"
      and eps: "rs_\<alpha> qsr2 = {q'. (qr,q') \<in> (ta_eps TA')^*}"
      and rrule: "(f qsr1 \<rightarrow> qr) \<in> ta_rules TA'" by auto
    from rule qs' have qs': "qs' = qsr1" by auto
    show "\<exists> q q'. (f qs' \<rightarrow> q) \<in> ta_rules TA' \<and> q' \<in> set qs \<and> (q,q') \<in> (ta_eps TA')^*"
      by (intro exI conjI, unfold qs', rule rrule, rule q, insert q2[unfolded rule] eps, auto)
  qed
qed
end

datatype ('q,'f)tree_automaton = Tree_Automaton "'q list" "('q,'f)ta_rule list" "('q \<times> 'q)list"
datatype 'q ta_relation = Id_Relation | Some_Relation "('q \<times> 'q) list"

fun relation_of :: "'q ta_relation \<Rightarrow> 'q rel" where
  "relation_of (Some_Relation rel) = set rel"
| "relation_of Id_Relation = Id"

fun rel_checker :: "'q ta_relation \<Rightarrow> ('q :: linorder) rs \<Rightarrow> 'q rs \<Rightarrow> 'q option" where 
  "rel_checker (Some_Relation rel) = (\<lambda> lhs rhs. let rlist = rs_to_list rhs in case (check_allm 
     (\<lambda> l. check (\<exists> r \<in> set rlist. (l,r) \<in> set rel) l) (rs_to_list lhs)) of Inl l \<Rightarrow> Some l | Inr _ \<Rightarrow> None)"
| "rel_checker Id_Relation = rs_subset"

lemma rel_checker:
  assumes "rel_checker rel lhs rhs = None" 
  shows "rs_\<alpha> lhs \<subseteq> (relation_of rel)^-1 `` (rs_\<alpha> rhs)"
proof (cases rel)
  case (Some_Relation rell)
  {
    assume "forallM (\<lambda>l. check (\<exists>r\<in>rs_\<alpha> rhs. (l, r) \<in> set rell) l) (rs_to_list lhs) = Inr ()" (is "?l = _")
    hence "isOK(?l)" by (cases ?l, auto)
    from this[simplified] have ?thesis by (auto simp: rs_correct Some_Relation)
  }
  with assms show ?thesis unfolding Some_Relation  
    by (auto split: option.splits sum.splits simp: rs_correct)
next
  case Id_Relation
  with assms show ?thesis by auto
qed

fun ta_of_ta :: "('q,'f)tree_automaton \<Rightarrow> ('q,'f)ta"
  where "ta_of_ta (Tree_Automaton fin rules eps) = 
  \<lparr> ta_final = set fin,
    ta_rules = set rules,
    ta_eps   = set eps
  \<rparr>"

fun check_coherent_rule :: "('q :: {show,linorder} \<Rightarrow> 'q list) \<Rightarrow> ('q \<times> 'q)rs \<Rightarrow> ('q,'f :: show)ta_rule list \<Rightarrow> ('q,'f)ta_rule \<Rightarrow> shows check" where
  "check_coherent_rule iter rel rules (TA_rule f qs q)  = 
      check_allm (\<lambda> i. 
        let qi = qs ! i in 
        check_allm (\<lambda> qi'. 
          let qs' = qs[ i := qi'] in
          check (filter (\<lambda> rule. case rule of 
            TA_rule g qs'' q' \<Rightarrow> f = g \<and> qs' = qs'' \<and> rs_memb (q,q') rel) rules \<noteq> [] ) 
            (shows ''rule '' +@+ shows f +@+ shows ''('' +@+ shows qs +@+ shows '') -> '' +@+ shows q +@+ 
             shows '' with '' +@+ shows (Suc i) +@+ shows ''. argument decreased to '' +@+ shows qi' +@+
             shows '' has no counterpart'') 
        ) (iter qi)
      ) [0 ..< length qs]"


fun check_coherent :: "('q :: {linorder,show},'f :: show)tree_automaton \<Rightarrow> 'q ta_relation \<Rightarrow> shows check" where
  "check_coherent _ Id_Relation = succeed"
| "check_coherent (Tree_Automaton fin rules _) (Some_Relation rel) = do {
    let iter = (\<lambda> q. map snd (filter (\<lambda> (a,b). a = q) rel));
    let rs = list_to_rs rel;
    check_subseteq (concat (map iter fin)) fin
      <+? (\<lambda> q. shows q +@+ shows '' is in relation to a final state, but not a final state itself'');
    check_allm (check_coherent_rule iter rs rules) rules
  }"

lemma check_coherent[simp]: "isOK(check_coherent TA rel) = (state_coherent (ta_of_ta TA) (relation_of rel))" (is "?l = ?r")
proof (cases rel)
  case Id_Relation
  with state_coherent_Id show ?thesis by auto
next
  case (Some_Relation rel) note rel = this
  obtain fin rules eps where ta: "TA = Tree_Automaton fin rules eps" by (cases TA, auto)
  let ?TA = "\<lparr> ta_final = set fin, ta_rules = set rules, ta_eps = set eps \<rparr>"
  let ?iter = "\<lambda> q. map snd (filter (\<lambda> (a,b). a = q) rel)"
  let ?rel = "list_to_rs rel"
  have TA: "ta_of_ta TA = ?TA" unfolding ta by simp
  have l: "?l = (set (concat (map ?iter fin)) \<subseteq> set fin \<and>
     (\<forall>x\<in>set rules. isOK (check_coherent_rule ?iter ?rel rules x)))" (is "_ = (?l1 \<subseteq> _ \<and> ?l2)")
     unfolding ta rel by (simp add: Let_def)
  have r: "?r = (set rel `` set fin \<subseteq> set fin \<and>
     (\<forall>f qs q. (f qs \<rightarrow> q) \<in> set rules \<longrightarrow>
               (\<forall>i<length qs. \<forall>qi. (qs ! i, qi) \<in> set rel \<longrightarrow> (\<exists>q'. (f qs[i := qi] \<rightarrow> q') \<in> set rules \<and> (q, q') \<in> set rel))))" 
    (is "_ = (?r1 \<subseteq> _ \<and> ?r2)")
    unfolding TA rel state_coherent_def by simp
  have id1: "?l1 = ?r1" by force
  have id2: "?l2 = ?r2" 
  proof
    assume ?l2
    show ?r2
    proof (intro allI impI)
      fix f qs q i qi
      assume rule: "(f qs \<rightarrow> q) \<in> set rules"
        and i: "i < length qs"
        and rel: "(qs ! i, qi) \<in> set rel"
      from i have i': "i \<in> {0 ..< length qs}" by simp
      let ?f = "(ta_rule_case (\<lambda>g qs'' q'. f = g \<and> qs[i := qi] = qs'' \<and> rs_memb (q, q') ?rel))"
      from `?l2`[rule_format, OF rule] have ok: "isOK (check_coherent_rule ?iter ?rel rules (f qs \<rightarrow> q))" .
      note ok = ok[simplified, rule_format, OF i' rel]
      from ok have "filter ?f rules \<noteq> []" by auto
      then obtain rule where f: "rule \<in> set rules \<and> ?f rule" by (cases "filter ?f rules", force+)
      thus "\<exists>q'. (f qs[i := qi] \<rightarrow> q') \<in> set rules \<and> (q, q') \<in> set rel" by (cases rule, auto simp: rs_correct)
    qed
  next
    assume ?r2
    show ?l2
    proof
      fix rule
      assume mem: "rule \<in> set rules"
      obtain f qs q where rule: "rule = (f qs \<rightarrow> q)" (is "_ = ?rule") by (cases rule, auto)
      let ?f = "\<lambda> i qi. ta_rule_case (\<lambda>g qs'' q'. f = g \<and> qs[i := qi] = qs'' \<and> (q, q') \<in> set rel)"
      have "isOK(check_coherent_rule ?iter ?rel rules ?rule) = (\<forall>i < length qs. 
        \<forall>qi. (qs ! i, qi) \<in> set rel \<longrightarrow>
            filter (?f i qi) rules \<noteq> [])" 
        by (auto simp: rs_correct)
      also have "..."
      proof (intro allI impI)
        fix i qi
        assume i: "i < length qs"
        assume rel: "(qs ! i, qi) \<in> set rel"
        from `?r2`[rule_format, OF mem[unfolded rule] i rel]
        obtain q' where "(f qs[i := qi] \<rightarrow> q') \<in> set rules" "(q, q') \<in> set rel" by auto
        hence "(f qs[i := qi] \<rightarrow> q') \<in> set (filter (?f i qi) rules)" by auto
        thus "filter (?f i qi) rules \<noteq> []" by force
      qed
      finally
      show "isOK(check_coherent_rule ?iter ?rel rules rule)" unfolding rule .
    qed
  qed
  show ?thesis unfolding l r id1 id2 ..
qed

fun is_Id_relation :: "'q ta_relation \<Rightarrow> bool" where
  "is_Id_relation Id_Relation = True"
| "is_Id_relation _ = False"

lemma is_Id_relation: "is_Id_relation rel \<Longrightarrow> relation_of rel = Id"
  by (cases rel, auto)

fun check_det :: "('q,'f)tree_automaton \<Rightarrow> shows check" where
  "check_det (Tree_Automaton fin rules eps) = (do {
      check (eps = []) (shows ''epsilon transitions not allowed'');
      check (distinct (map (\<lambda> rule. case rule of TA_rule f qs q \<Rightarrow> (f,qs)) rules)) (shows ''some lhs occurs twice'')
    }) <+? (\<lambda> s. shows ''problem when ensuring determinism of automata'' +@+ shows_nl +@+ s)"

lemma check_det: assumes ok: "isOK(check_det TA)"
  shows "ta_det (ta_of_ta TA)"
proof (cases TA)
  case (Tree_Automaton fin rules eps) note ta = this
  note ok = ok[unfolded ta, simplified]
  let ?ta = "\<lparr> ta_final = set fin, ta_rules = set rules, ta_eps = {} \<rparr>"
  from ok have ta: "ta_of_ta TA = ?ta" unfolding ta by auto
  from ok have det: "distinct (map (ta_rule_case (\<lambda>f qs q. (f, qs))) rules)" by auto
  show ?thesis unfolding ta ta_det_def
  proof (rule conjI, simp, intro allI impI, simp)
    fix f qs q q'
    assume q: "(f qs \<rightarrow> q) \<in> set rules" and q': "(f qs \<rightarrow> q') \<in> set rules" 
    from det[unfolded distinct_map inj_on_def, THEN conjunct2,  
     THEN bspec[OF _ q], THEN bspec[OF _ q']] 
    show "q = q'" by auto
  qed
qed

fun generate_ta :: "('q :: linorder,'f :: linorder)tree_automaton \<Rightarrow> ('q,'f)ta_impl"
  where "generate_ta (Tree_Automaton fin rules eps) = (let ep = memo_rbt_rtrancl eps;
     epr = memo_rbt_rtrancl (map (\<lambda> (q,q'). (q',q)) eps);
     rqs_rs = rs_Union (map (\<lambda> rule. (ep (r_rhs rule))) rules);
     rrules = map (conv_ta_rule ep) rules
     in 
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm r_sym_impl rrules,
         ta_r_lhs_states_impl = rs_to_list rqs_rs,
         ta_rhs_states_set = rqs_rs,
         ta_eps_impl = eps,
         ta_epss_impl = ep,
         ta_epsrs_impl = epr
          \<rparr> )"

definition generate_ta_cond :: "('q :: {linorder, show},'f :: {linorder,show})tree_automaton \<Rightarrow> 'q ta_relation \<Rightarrow> shows + ('q,'f)ta_impl"
  where "generate_ta_cond ta rel \<equiv> case ta of (Tree_Automaton _ _ eps) \<Rightarrow> do
    {
      check (is_Id_relation rel \<or> eps = []) (shows ''relation must be Id, or no epsilon rules must be present'');
      check_coherent ta rel <+? (\<lambda> s. shows ''automaton is not coherent w.r.t. relation'' +@+ shows_nl +@+ s);
      return (generate_ta ta)
    }" 

lemma generate_ta_rules: "ta_rules (ta_of (generate_ta (Tree_Automaton fin rules eps))) = set rules"
  (is "ta_rules ?TA' = _")
proof -
  {
    fix eps
    let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). r_sym_impl d = fn} | fn. True}"
    {
      fix r
      assume "r \<in> ?rules"
      from this[simplified]
      obtain rl rs where r: "r = ta_rule_conv rl"
        and          rl: "rl = conv_ta_rule eps rs"
        and           rs: "rs \<in> set rules"
        by auto
      hence "r \<in> set rules" by (cases rs, auto)
    }
    moreover
    {
      fix r 
      assume "r \<in> set rules"
      then obtain f qs q where r: "r = (f qs \<rightarrow> q)" and rr: "(f qs \<rightarrow> q) \<in> set rules" by (cases r, auto)
      have "conv_ta_rule eps r = TA_rule_impl f qs q (eps q)" (is "_ = ?rl") unfolding r  by simp
      with rr have rl: "?rl \<in> set (map (conv_ta_rule eps) rules)" (is "_ \<in> ?L") by force
      have r: "r = ta_rule_conv ?rl" unfolding r by auto
      have "r \<in> ta_rule_conv ` ?L" unfolding r using rl by blast
      hence "r \<in> ?rules" by blast
    }
    ultimately have "?rules = set rules" by blast
  }
  thus ?thesis
    unfolding generate_ta.simps  Let_def
    unfolding ta_of_def ta.simps ta_impl.simps elem_list_to_rm.rm_set_lookup
    by auto
qed

  
lemma generate_ta_cond: assumes cond: "generate_ta_cond TA rel = return ta"
  shows "ta_inv ta (rel_checker rel) (relation_of rel) \<and> ta = generate_ta TA"
proof (cases TA)
  case (Tree_Automaton fin rules epsilon)
  note cond = cond[unfolded generate_ta_cond_def Tree_Automaton] 
  from cond have choice: "is_Id_relation rel \<or> epsilon = []" by auto
  from cond have ta: "ta = generate_ta (Tree_Automaton fin rules epsilon)" by auto
  from choice is_Id_relation[of rel] have choice: "relation_of rel = Id \<or> set epsilon = {}" by auto
  obtain eps where eps: "eps = memo_rbt_rtrancl epsilon" by auto
  obtain epsr where epsr: "epsr = memo_rbt_rtrancl (map (\<lambda>(q,q'). (q',q)) epsilon)" by auto
  let ?rqs = "rs_Union (map (\<lambda> rule. (eps (r_rhs rule))) rules)"
  let ?rrules = "map (conv_ta_rule eps) rules"
  from ta eps epsr have A: "ta = 
       \<lparr> ta_final_impl = list_to_rs fin, 
         ta_rules_impl = elem_list_to_rm r_sym_impl ?rrules,
         ta_r_lhs_states_impl = rs_to_list ?rqs,
         ta_rhs_states_set = ?rqs,
         ta_eps_impl = epsilon,
         ta_epss_impl = eps,
         ta_epsrs_impl = epsr
          \<rparr>" (is "_ = ?TA")
    by (auto simp: Let_def)
  let ?TA' = "ta_of ?TA"
  let ?rules = "\<Union> { ta_rule_conv ` {d \<in> set (map (conv_ta_rule eps) rules). r_sym_impl d = fn} | fn. True}"
  have ta_rules: "ta_rules ?TA' = set rules"
    using generate_ta_rules[of fin rules epsilon] 
    by (simp add: eps epsr Let_def)
  have qq1: "{(q, q'). (q, q') \<in> (set epsilon)^*} = (set epsilon)^*" by auto
  have qq2: "((\<lambda>(q,q'). (q',q)) ` set epsilon)^* = ((set epsilon)^*)^-1" unfolding rtrancl_converse[symmetric]
    by (rule arg_cong[where f = "\<lambda> x. x^*"], auto)
  have one: "ta_rhs_states ?TA' = rs_\<alpha> (ta_rhs_states_set ?TA)" 
    unfolding ta_rhs_states_def ta_rules 
    unfolding ta_of_def ta.simps ta_impl.simps rs_Union
    by (simp add: eps memo_rbt_rtrancl qq1, auto)
  have two: "set (ta_r_lhs_states_impl ?TA) = ta_rhs_states ?TA'" unfolding one 
    by (auto simp: rs_correct Let_def) 
  let ?ta_of = "\<lparr>ta_final = set fin, ta_rules = set rules, ta_eps = set epsilon\<rparr>"
  from cond have coh: "state_coherent ?ta_of (relation_of rel)" by auto 
  have id: "?ta_of = ta_of ?TA" using ta_rules unfolding ta_of_def 
    by (simp add: rs_correct)
  with coh have coh: "state_coherent (ta_of ?TA) (relation_of rel)" by simp
  have "ta_inv ta (rel_checker rel) (relation_of rel)" unfolding A
    by (unfold_locales, insert coh, unfold one two, unfold ta_impl.simps elem_list_to_rm.rm_set_lookup ta_rules, 
    insert choice, auto simp: ta_of_def eps epsr memo_rbt_rtrancl qq1 qq2 rel_checker rs_correct) 
  thus ?thesis unfolding Tree_Automaton ta by auto
qed

definition tree_aut_trs_compat :: "('q :: {linorder,show},'f :: {linorder,show})tree_automaton \<Rightarrow> 'q ta_relation \<Rightarrow> ('f,'q)rules \<Rightarrow> shows check"
  where "tree_aut_trs_compat ta rel R \<equiv> do {
     TA \<leftarrow> generate_ta_cond ta rel;
     check_wf_trs R;
     (if isOK (check_left_linear_trs R) then succeed else
       check_det ta <+? (\<lambda> s. shows ''could not ensure left-linearity of determinism'' +@+ shows_nl +@+ s));
     state_compatible_eff_list TA (rel_checker rel) R
        <+? (\<lambda> (lr,lrq,q). shows ''TA is not compatible with R'' +@+ shows_nl 
             +@+ shows ''for rule '' +@+ shows_rule lr +@+ shows_nl
             +@+ shows ''which is instantiated by states to '' +@+ shows_rule lrq +@+ shows_nl
             +@+ shows ''the state '' +@+ shows q +@+ shows '' is only reachable from the lhs'' +@+ shows_nl)
  }"

lemma tree_aut_trs_compat:
  assumes ok: "isOK(tree_aut_trs_compat ta rel R)"
  shows "{t | s t. s \<in> ta_lang (ta_of_ta ta) \<and> (s,t) \<in> (rstep (set R))^*} \<subseteq> ta_lang (ta_of_ta ta)"
proof (cases ta)
  case (Tree_Automaton fin rules eps) note ta = this
  note ok = ok[unfolded tree_aut_trs_compat_def Let_def]
  from ok obtain TA where gc: "generate_ta_cond ta rel = return TA" 
    by auto
  let ?ta = "generate_ta ta"  
  from generate_ta_cond[OF gc] have TA: "TA = ?ta" and inv: "ta_inv ?ta (rel_checker rel) (relation_of rel)" by auto
  note ok = ok[unfolded gc TA, simplified]
  from ok 
  have wf: "wf_trs (set R)" and 
    compat: "isOK( state_compatible_eff_list ?ta (rel_checker rel) R)" by auto
  let ?TA = "\<lparr>ta_final = set fin, ta_rules = set rules, ta_eps = set eps\<rparr>"
  let ?ll = "left_linear_trs (set R)"
  from ok have "?ll \<or> isOK(check_det ta)" 
    by (auto split: if_splits)
  with check_det have "?ll \<or> ta_det (ta_of_ta ta)" by auto
  with ta have ll: "?ll \<or> ta_det ?TA" by auto
  {
    fix l r
    assume "(l,r) \<in> set R"
    hence "vars_term r \<subseteq> vars_term l" using wf  unfolding wf_trs_def by auto
  } note var_cond = this
  interpret ta_inv "generate_ta ta" "rel_checker rel" "relation_of rel" by fact 
  show ?thesis unfolding ta ta_of_ta.simps
  proof (rule state_compatible[OF _ _ _ var_cond subset_refl])
    let ?TA = "\<lparr>ta_final = set fin, ta_rules = set rules, ta_eps = set eps\<rparr>"
    let ?T = "ta_of (generate_ta ta)"    
    have one: "ta_rules ?T = set rules" unfolding ta using generate_ta_rules[of fin rules eps] by auto
    have two: "ta_final ?T = set fin" unfolding ta by (simp add: Let_def ta_of_def rs_correct)
    have three: "ta_eps ?T = set eps" unfolding ta by (simp add: Let_def ta_of_def)
    from one two three have id: "?T = ?TA" by auto
    show "state_compatible ?TA (relation_of rel) (set R)" 
      by (rule state_compatible_eff_list[OF var_cond compat, unfolded id])
    from no_eps_or_rel_Id
    show "state_coherent ?TA (relation_of rel)" using id coherent by auto
  qed (insert no_eps_or_rel_Id ll, auto simp: ta ta_of_def Let_def)
qed

datatype ('q,'f)tree_aut_problem = tree_aut_problem "('q,'f)tree_automaton" "('f,'q)rules" "('q,'f)tree_automaton"

fun certify_tree_aut_problem :: "('q :: {linorder,show},'f :: {linorder,show})tree_aut_problem \<Rightarrow> shows check"
  where "certify_tree_aut_problem (tree_aut_problem A R B) =
    tree_aut_trs_compat B Id_Relation R"

end

